package WebService::Deltacloud;

use warnings;
use strict;

BEGIN {

    use version 0.77; our( $VERSION ) = version->declare( "0.0.1" );

    use Carp;
    use Data::Dumper;
    use Readonly;

    use Moose;
    use Moose::Util::TypeConstraints;
    use MooseX::Params::Validate;
    use MooseX::Types::Moose qw( Str HashRef );
    use MooseX::Types::URI qw( Uri );

    use HTTP::Request;
    use URI::QueryParam;
    use URI::Split qw( uri_split uri_join );

    require LWP::UserAgent;
    require XML::Simple;

    use WebService::Deltacloud::Types qw( :all );

    our( @ISA, @EXPORT, @EXPORT_OK );
    @ISA = qw( Exporter );
    @EXPORT = qw();
    @EXPORT_OK = qw( hardware_profiles realms images instances );
};

=head1 NAME

WebService::Deltacloud - The great new WebService::Deltacloud!

=head1 VERSION

Version 0.0.1

=cut

our $VERSION = '0.01';

Readonly my $APIDEFAULT => 'http://localhost:3001/api';
our( $APIBASE ) = $APIDEFAULT;

=head1 SYNOPSIS

Quick summary of what the module does.

Perhaps a little code snippet.

    use WebService::Deltacloud;

    my $foo = WebService::Deltacloud->new();
    ...

=head1 EXPORT

A list of functions that can be exported.  You can delete this section
if you don't export anything, such as for a purely object-oriented module.

=cut

### PRIVATE

# LWP::UserAgent
has '_ua' => (
    is          => 'ro',
    isa         => UserAgent,
    coerce      => 1,
    lazy_build  => 1,
    required    => 1,
);

sub _build__ua { LWP::UserAgent->new };

# XML::Simple
has '_xml' => (
    is          => 'rw',
    isa         => XMLSimple,
    coerce      => 1,
    lazy_build  => 1,
    required    => 1,
);

sub _build__xml { 
    # set some default XML::Simple options, overridable
    my( %options ) = ( 
        KeyAttr     => [],
        ForceArray  => 1,
    );

    # add options if provided
    defined( $_ ) and %options = ( %options, %{$_} );

    XML::Simple->new( %options );
};

sub _compose_uri {

    my( $self, %params ) = validated_hash( \@_, 
        base_uri => {
            isa         => Uri,
            coerce      => 1,
            optional    => 1,
        },
        target => {
            isa         => APITarget,
        },
        params => {
            isa         => HashRef,
            optional    => 1,
        },
    );

    defined( $params{base_uri} )
        or $params{base_uri} = $self->base_uri;

    # parse params
    my( $scheme, $auth, $path, $query, $frag ) = uri_split( $params{base_uri} );

    # add the target
    $path .= $params{target};

    # add query params if any are available
    my( $result ) = URI->new( uri_join( $scheme, $auth, $path, $query, $frag ) );
    foreach my $query_param ( keys( %{$params{params}} ) ) {
        $result->query_param( $query_param => $params{params}->{$query_param} );
    }

    return( $result );
}

sub _api_request {

    my( $self, %params ) = validated_hash( \@_, 
        method  => {
            isa     => HTTPMethod,
        },
        ua      => {
            isa         => UserAgent,
            coerce      => 1,
            optional    => 1,
        },
        uri     => {
            isa         => Uri,
            coerce      => 1,
        },
        authenticate => {
            isa         => 'Bool',
            coerce      => 1,
            default     => 0,
        },
    );

    defined( $params{ua} ) or $params{ua} = $self->_ua;

    my( $request ) = HTTP::Request->new( $params{method} => $params{uri} );

    # add credentials if provided
    if ( $params{authenticate} ) {

        my( $username ) = $self->credentials->access_key_id;
        my( $password ) = $self->credentials->secret_access_key;
        $request->authorization_basic( $username, $password );
    }

    # set encoding if necessary
    if ( $params{method} eq 'POST' ) {
    
        $request->content_type( 'application/x-www-form-encoded' );
    }

    print Data::Dumper->Dump( [$request], [qw(*request)] );

    my( $response ) = $params{ua}->request( $request ); 

    return( $response );
}

sub _parse_content {

    my( $self, %params ) = validated_hash( \@_, 
        response  => {
            isa         => 'HTTP::Response',
            required    => 1,
        },
        keys => {
            isa         => HashRef,
            optional    => 1,
        },
        values => {
            isa         => HashRef,
            optional    => 1,
        },
        group => {
            isa         => HashRef,
            optional    => 1,
        },
    );

    my( $xmloptions ) = {
        ForceArray  => [ keys( %{$params{keys}} ) ],
        KeyAttr     => $params{keys},
        ValueAttr   => $params{values},
        GroupTags   => $params{group},
        SuppressEmpty   => 1,
    };

    my( $result );

    if ( $params{response}->is_success ) {

        my( $raw ) = $params{response}->decoded_content;
        print Data::Dumper->Dump( [$xmloptions], [qw(*xmloptions)] );
        $result = $self->_xml->XMLin( $raw, %{$xmloptions} );
    }
    else {
        $result->{error} = {
            status  => $params{response}->status_line,
            text    => $params{response}->as_string,
            html    => $params{response}->error_as_HTML,
        };
    }

    return( $result );
}

sub _process_api {

    my( $self, %params ) = validated_hash( \@_, 
        target => {
            isa     => APITarget,
        },
        params => {
            isa     => HashRef,
            default => {},
        },
        method => {
            isa     => HTTPMethod,
            default => 'GET',
        },
        keys => {
            isa     => HashRef,
            default => {},
        },
        values => {
            isa     => HashRef,
            default => {},
        },
    );

    # instantiate the URI object
    my( $base_uri ) = $self->base_uri;

    my( $uri ) = $self->_compose_uri(
        base_uri    => $base_uri,
        target      => $params{target},
        params      => $params{params},
    );

    my( $response ) = $self->_api_request(
        method  => $params{method},
        uri     => $uri,
    );

    # do we need to authenticate?
    if ( ( $response->is_error ) && ( $response->code == 401 ) ) {
     
        $response = $self->_api_request(
            method          => $params{method},
            uri             => $uri,
            authenticate    => 1,
        );
    }

    my( $parsed ) = $self->_parse_content(
        response    => $response,
        keys        => $params{keys},
        values      => $params{values},
    );

    return( $parsed );
}

### PUBLIC

# base URI
has 'base_uri' => (
    is          => 'ro',
    isa         => Uri,
    coerce      => 1,
    lazy_build  => 1,
    required    => 1,
);

sub _build_base_uri { URI->new( $APIBASE ) };

# credentials
has 'credentials' => (
    is          => 'rw',
    isa         => AWSProfile,
    coerce      => 1,
    lazy_build  => 1,
);

sub hardware_profiles {

    my( $self, %params ) = validated_hash( \@_, 
        id => {
            isa         => Str,
            optional    => 1,
        },
    );

    my( $keys ) = {
        hardware_profile    => 'id',
        property            => 'name',
        enum                => 'value',
    };

    my( $values ) = {
        hardware_profile    => 'name',
    };

    my( $target ) = '/hardware_profiles';
    my( $queryparams ) = {};

    if ( defined( $params{id} ) ) {

        $target .= join( '', '/', $params{id} );
        $queryparams->{id} = $params{id};
    }

    my( $result ) = $self->_process_api(
        target  => $target,
        params  => $queryparams,
        method  => 'GET',
        keys    => $keys,
        values  => $values,
    );

    if ( defined( $result->{error} ) ) {

        # that's no good
        carp( $result->{error}->{status} );
        return( $result );
    }
    else {

        # that's good
        return( $result );
    }
}

sub realms {

    my( $self, %params ) = validated_hash( \@_, 
        id => {
            isa         => Str,
            optional    => 1,
        },
    );

    my( $keys ) = {
        realm       => 'id',
    };

    my( $values ) = {
        realm => 'name',
        realm => 'state',
    };

    my( $target ) = '/realms';
    my( $queryparams ) = {};

    if ( defined( $params{id} ) ) {

        $target .= join( '', '/', $params{id} );
        $queryparams->{id} = $params{id};
    }

    my( $result ) = $self->_process_api(
        target  => $target,
        params  => $queryparams,
        method  => 'GET',
        keys    => $keys,
        values  => $values,
    );

    if ( defined( $result->{error} ) ) {

        # that's no good
        croak( $result->{error}->{status} );
    }
    else {

        # that's good
        return( $result );
    }
}

sub images {

    my( $self, %params ) = validated_hash( \@_, 
        id => {
            isa         => Str,
            optional    => 1,
        },
    );

    my( $keys ) = {
        image       => 'id',
    };

    my( $values ) = {
        image => 'name',
        image => 'owner',
        image => 'description',
        image => 'architecture',
    };

    my( $target ) = '/images';
    my( $queryparams ) = {};

    if ( defined( $params{id} ) ) {

        $target .= join( '', '/', $params{id} );
        $queryparams->{id} = $params{id};
    }

    my( $result ) = $self->_process_api(
        target      => $target,
        params      => $queryparams,
        method      => 'GET',
        keys        => $keys,
        values      => $values,
    );

    if ( defined( $result->{error} ) ) {

        # that's no good
        croak( $result->{error}->{status} );
    }
    else {

        # that's good
        return( $result );
    }
}

sub instances {

    my( $self, %params ) = validated_hash( \@_, 
        id => {
            isa         => Str,
            optional    => 1,
        },
        image => {
            isa         => Str,
            optional    => 1,
        },
        realm => {
            isa         => Str,
            optional    => 1,
        },
        profile => {
            isa         => Str,
            optional    => 1,
        },
        name => {
            isa         => Str,
            optional    => 1,
        },
        key => {
            isa         => Str,
            optional    => 1,
        },
        loadbalancer => {
            isa         => Str,
            optional    => 1,
        },
    );

    my( $keys ) = {
        instance            => 'id',
        link                => 'rel',
        image               => 'href',
        hardware_profile    => 'href',
        realm               => 'href',
        public_addresses    => 'address',
        private_addresses   => 'address',
    };

    my( $values ) = {
        instance => 'name',
        instance => 'owner',
        instance => 'description',
        instance => 'architecture',
        instance => 'state',
    };

    my( $group ) = {
    };

    my( $target ) = '/instances';
    my( $method ) = 'GET';
    my( $queryparams ) = {};

    if ( defined( $params{id} ) ) {

        $target .= join( '', '/', $params{id} );
        $queryparams->{id} = $params{id};
    }
    elsif ( defined( $params{image} ) ) {

        # we're creating a new instance
        $queryparams->{image_id} = $params{image};
        defined( $params{realm} ) && ( $queryparams->{realm_id} = $params{realm} );
        defined( $params{profile} ) && ( $queryparams->{hwp_id} = $params{profile} );
        defined( $params{name} ) && ( $queryparams->{name} = $params{name} );
        defined( $params{key} ) && ( $queryparams->{keyname} = $params{key} );
        defined( $params{loadbalancer} ) && ( $queryparams->{load_balancer_id} = $params{loadbalancer} );

        $method = 'POST';
    }

    my( $result ) = $self->_process_api(
        target  => $target,
        params  => $queryparams,
        method  => $method,
        keys    => $keys,
        values  => $values,
    );

    if ( defined( $result->{error} ) ) {

        # that's no good
        croak( $result->{error}->{status} );
    }
    else {

        # that's good
        return( $result );
    }
}

=head1 AUTHOR

Steve Huff, C<< <shuff at cpan.org> >>

=head1 BUGS

Please report any bugs or feature requests to C<bug-webservice-deltacloud at rt.cpan.org>, or through
the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=WebService-Deltacloud>.  I will be notified, and then you'll
automatically be notified of progress on your bug as I make changes.




=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc WebService::Deltacloud


You can also look for information at:

=over 4

=item * RT: CPAN's request tracker

L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=WebService-Deltacloud>

=item * AnnoCPAN: Annotated CPAN documentation

L<http://annocpan.org/dist/WebService-Deltacloud>

=item * CPAN Ratings

L<http://cpanratings.perl.org/d/WebService-Deltacloud>

=item * Search CPAN

L<http://search.cpan.org/dist/WebService-Deltacloud/>

=back


=head1 ACKNOWLEDGEMENTS


=head1 LICENSE AND COPYRIGHT

Copyright 2011 Steve Huff.

This program is free software; you can redistribute it and/or modify it
under the terms of either: the GNU General Public License as published
by the Free Software Foundation; or the Artistic License.

See http://dev.perl.org/licenses/ for more information.


=cut

no Moose;
__PACKAGE__->meta->make_immutable;
1; # End of WebService::Deltacloud
